﻿using Asp.Versioning;
using IdentityManagerUI.Models;
using InfProject.Controllers;
using InfProject.DBContext;
using InfProject.Models;
using InfProject.Utils;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Identity;
using Microsoft.AspNetCore.Mvc;
using Microsoft.CodeAnalysis.CSharp.Syntax;
using Newtonsoft.Json;
using System;
using System.Data;
using System.Net.Mime;
using System.Security.Claims;
using TypeValidationLibrary;
using WebUtilsLib;

namespace InfProject.Areas.VroApi.Controllers
{
    [Area("DeviceApi")]
    [ApiVersion(1.0)]
    [ApiController]
    //[Route("deviceapi/v{v:apiVersion}")]
    [Route("deviceapi/v1")]
    public class DeviceApiController : ControllerBase
    {
        // as a suggestion to generate values: https://www.uuidgenerator.net/version4
        // Claim name to use Vro Api and Api key name
        public const string ApiKeyName = "DeviceApi";  // ClaimType в AspNetUserClaims
        public string Version => ((ApiVersionAttribute)Attribute.GetCustomAttribute(typeof(DeviceApiController), typeof(ApiVersionAttribute))).Versions[0].ToString();
        public string Get() => $"Device (user agnostic) API on MatInf database. Version {Version}";


        private readonly ILogger<DeviceApiController> logger;
        private readonly IConfiguration config;
        private readonly DataContext dataContext;
        private readonly IHttpContextAccessor httpContextAccessor;

        public DeviceApiController(ILogger<DeviceApiController> logger, IConfiguration config, DataContext dataContext, IHttpContextAccessor httpContextAccessor)
        {
            this.logger = logger;
            this.config = config;
            this.dataContext = dataContext;
            this.httpContextAccessor = httpContextAccessor;
        }

        private string connectionString;
        private string ip;
        private int deviceUserId;

        /// <summary>
        /// 1) initialize appropriate connection string to use
        /// 2) Check API Security
        /// 3) defide deviceUserId
        /// </summary>
        /// <returns>true - ok; false - no access</returns>
        private async Task<bool> InitApi()
        {
            var connectionString = dataContext.GetConnectionString(config, httpContextAccessor);
            string apiKeyValue = Request.Headers[ApiKeyName];
            ip = Request.HttpContext.Connection.RemoteIpAddress.ToString();
            deviceUserId = await dataContext.GetActiveUserIdWithClaimValue(ApiKeyName, apiKeyValue);
            if (deviceUserId == 0) {
                logger.LogError("Authorisation failed for user provided claimValue {apiKeyValue} [ip={ip}]", apiKeyValue, ip);
            }
            return deviceUserId != 0;
        }


        /// <summary>
        /// return type list that could be used within a tenant
        /// </summary>
        /// <returns>JsonResult - list with types</returns>
        [HttpPost("gettypes")]
        //[Consumes(MediaTypeNames.Application.Octet)] // application/octet-stream
        [Produces(typeof(JsonResult))]
        public async Task<ActionResult> GetTypes()
        {
            if (!await InitApi())
            {  // get connectionString and database Name
                return Unauthorized();
            }
            List<TypeInfo> types = await dataContext.GetTypes();
            logger.LogInformation("GetTypes() returned {count} item(s). [deviceUserId={deviceUserId}, ip={ip}]", types.Count, deviceUserId, ip);
            string json = JsonConvert.SerializeObject(types);
            return Content(json, "application/json");
        }


        /// <summary>
        /// return user list
        /// </summary>
        /// <returns>JsonResult - list with users</returns>
        [HttpPost("getusers")]
        //[Consumes(MediaTypeNames.Application.Octet)] // application/octet-stream
        [Produces(typeof(JsonResult))]
        public async Task<ActionResult> GetUsers()
        {
            if (!await InitApi())
            {  // get connectionString and database Name
                return Unauthorized();
            }
            List<WebAppUser> users = await dataContext.GetUserListActive();
            logger.LogInformation("GetUsers() returned {count} item(s). [deviceUserId={deviceUserId}, ip={ip}]", users.Count, deviceUserId, ip);
            string json = JsonConvert.SerializeObject(users);
            return Content(json, "application/json");
        }




        /// <summary>
        /// return type list that could be used within a tenant
        /// </summary>
        /// <returns>JsonResult - list with types</returns>
        [HttpPost("getrubrics")]
        //[Consumes(MediaTypeNames.Application.Octet)] // application/octet-stream
        [Produces(typeof(JsonResult))]
        public async Task<ActionResult> GetRubrics(int typeId, int maxLevel = 10)
        {
            if (!await InitApi())
            {  // get connectionString and database Name
                return Unauthorized();
            }
            //List<RubricInfo> rubrics = await dataContext.GetList_RubricTree(typeId);
            AccessControlFilter accessFilter = AccessControlFilter.SuperAdmin; // HttpContext.GetAccessControlFilter();
            List<RubricInfo> rubrics = await dataContext.GetList_RubricTree_AccessControl(typeId, accessFilter, maxLevel);
            logger.LogInformation("GetRubrics({typeId}) returned {count} item(s). [deviceUserId={deviceUserId}, ip={ip}]", typeId, rubrics.Count, deviceUserId, ip);
            string json = JsonConvert.SerializeObject(rubrics);
            return Content(json, "application/json");
        }



        /// <summary>
        /// Searches for objects/samples by applying filter criteria to tenant's data
        /// </summary>
        /// <param name="searchMode">GetElements_Filter | GetObjectInfo_Filter | GetSample_Filter</param>
        /// <param name="filter">filter with search criteria</param>
        /// <returns>JsonResult - search result (object/samples list)</returns>
        [HttpPost("searchbyfilter")]
        //[Consumes(MediaTypeNames.Application.Octet)] // application/octet-stream
        [Produces(typeof(JsonResult))]
        public async Task<ActionResult> SearchByFilter(SearchStoredProcedure_Filter searchMode, Filter filter)
        {
            if (!await InitApi())
            {  // get connectionString and database Name
                return Unauthorized();
            }
            int count = 0;
            List<ObjectInfo> listObjects = null!;
            List<Sample> listSamples = null!;

            InfProject.Models.TypeInfo? type = filter.TypeId != 0 ? await dataContext.GetType(filter.TypeId) : null;
            if (filter.HasChemicalSystem(type) || searchMode == SearchStoredProcedure_Filter.GetSample_Filter)
            {
                listSamples = await dataContext.SearchByFilter<Sample>(SearchStoredProcedure_Filter.GetSample_Filter, filter);
                count = listSamples.Count;
            }
            else {
                listObjects = await dataContext.SearchByFilter<ObjectInfo>(searchMode, filter);
                count = listObjects.Count;
            }
            logger.LogInformation("SearchByFilter({searchMode}, {filter}) returned {count} item(s). [deviceUserId={deviceUserId}, ip={ip}]", searchMode, filter, count, deviceUserId, ip);
            string json = JsonConvert.SerializeObject(listSamples != null ? listSamples : listObjects);
            return Content(json, "application/json");
        }


        /// <summary>
        /// Add object, described by json to the database
        /// </summary>
        /// <param name="json">JSON with an object to add</param>
        /// <returns>JsonResult - search result (object/samples list)</returns>
        [HttpPost("processjson")]
        //[Consumes(MediaTypeNames.Application.Octet)] // application/octet-stream
        [Produces(typeof(JsonResult))]
        public async Task<ActionResult> ProcessJson(string json/*, int[]? parentObjectId=null, int linkTypeObjectId=0*/)
        {
            // <param name="parentObjectId">if set (!=0) a link created to the parent object</param>
            // <param name="linkTypeObjectId">if set (!=0) a link created to the parent object with LinkTypeObjectId specified</param>

            if (!await InitApi())
            {  // get connectionString and database Name
                return Unauthorized();
            }

            DeviceApi.DeviceApiUpdater updater = new DeviceApi.DeviceApiUpdater(dataContext, logger, deviceUserId);
            string outputJson = await updater.ProcessJson(json);

            logger.LogInformation("ProcessJson({json}) returned {outputJson}", json, outputJson);
            return Content(json, "application/json");
        }






    }
}
