﻿using Microsoft.AspNetCore.Mvc;
using Microsoft.AspNetCore.Mvc.ViewEngines;
using System.Data.SqlClient;
using System.Data;
using System.Diagnostics;
using Dapper;
using System.Reflection;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Authorization;
using InfProject.DBContext;
using InfProject.Models;
using InfProject.Utils;
using WebUtilsLib;
using System.Linq;
using static System.Runtime.InteropServices.JavaScript.JSType;
using System.Collections.Generic;
using TypeValidationLibrary;

namespace InfProject.Controllers;

public class AjaxController : Controller
{
    private readonly ILogger<AjaxController> logger;
    private readonly IConfiguration config;
    private readonly DataContext dataContext;

    public AjaxController(ILogger<AjaxController> logger, IConfiguration config, DataContext dataContext)
    {
        this.logger = logger;
        this.config = config;
        this.dataContext = dataContext;
    }

    [HttpPost]
    public async Task<IActionResult> SearchObjectList([FromForm] string query, [FromForm] int typeId, [FromForm] int objectId)
    {
        var data = await dataContext.ObjectInfo_Find(query, typeId, objectId);
        return Json(data);
    }

    [HttpGet]
    [Authorize(Roles = "User,PowerUser,Administrator")]
    public async Task<IActionResult> GetUsers()
    {
        // value: item.id,
        // text: `${item.project}) ${item.name} [${item.email}]`
        var data = await dataContext.GetUserListActive();
        return Json(data.Select(x => new { id=x.Id, name=x.Name, email=x.Email, project=x.Project }).OrderBy(x => $"{x.project} {x.name}"));
    }


    [HttpGet]
    [Authorize(Roles = "User,PowerUser,Administrator")]
    public async Task<IActionResult> GetProjects()
    {
        // value: item.id,
        // text: `${ item.project}) ${ item.count} member(s)[${ item.names}]`
        IEnumerable<(string project, int countUsers, string namesCsv)> data = await dataContext.GetActiveProjectsExtended();
        return Json(data.Select(x => new { id = x.project, project = x.project, count = x.countUsers, namesCsv = x.namesCsv }).OrderBy(x => x.project));
    }


    [HttpGet]
    public async Task<IActionResult> GetRubricText([FromQuery] int rubricid)
    {
        // value: item.id,
        // text: `${item.project}) ${item.name} [${item.email}]`
        var text = await dataContext.GetRubricText(rubricid);
        return Json(new { text });
    }

    [HttpGet]
    public async Task<IActionResult> GetObjectLinkRubricByObject([FromQuery] int objectid)
    {
        IEnumerable<ObjectLinkRubric> list = await dataContext.GetObjectLinkRubricByObject(objectid);
        return Json(new { ObjectLinkRubric = list });
    }

    [HttpGet]
    public async Task<IActionResult> GetSortCode_ObjectLinkRubric([FromQuery] int objectid, [FromQuery] int rubricid)
    {
        var list = await dataContext.GetList_ExecDevelopment<ObjectLinkRubric>("select top 1 * FROM dbo.ObjectLinkRubric WHERE RubricId=@rubricid AND ObjectId=@objectid",
            new { rubricid, objectid });
        ObjectLinkRubric res = list.Count>0 ? list[0] : null;
        string error = string.Empty;
        if (res == null) {
            error = "404 - not found";
        }
        if (string.IsNullOrEmpty(error) && HttpContext.IsWriteDenied(AccessControl.Public, (int)res._createdBy)) {
            error = "access denied (other user created the link)";
        }
        return Json(new { ObjectLinkRubric = res, error });
    }

    /// <summary>
    /// Get Object Information as JSON object
    /// </summary>
    /// <param name="objectId">ObjectId - has a priority</param>
    /// <param name="url">ObjectUrl - used if ObjectId==0</param>
    /// <returns>JSON</returns>
    [HttpGet]
    public async Task<IActionResult> GetObject([FromQuery] int objectId, [FromQuery] string url)
    {
        // https://mdi.matinf.pro/ajax/getobject?objectid=1
        // https://mdi.matinf.pro/ajax/getobject?objectid=2
        ObjectInfo o;
        if (objectId != 0)
        {
            o = await dataContext.ObjectInfo_Get(objectId);
        }
        else
        {
            o = await dataContext.GetObjectByUrl(url);
        }
        if (o == null || o.ObjectId == 0)
        {
            return NotFound($"Can not find object ({objectId}, {url})");
        }
        if (HttpContext.IsReadDenied(o.AccessControl, (int)o._createdBy))
        {
            return this.RedirectToLoginOrAccessDenied(HttpContext);
        }
        var curType = await dataContext.GetType(o.TypeId);
        var dboName = InfProject.Models.TypeInfo.GetViewOrTableName(curType.TableName);
        // string sql = $@"select * from dbo.ObjectInfo WHERE ObjectId=@ObjectId FOR JSON AUTO";    // old ObjectInfo-only version
        // string sql = $@"select * from {dboName} WHERE ObjectId=@ObjectId FOR JSON AUTO";    // trims JSON output to ~ 2K chars
        string sql = $@"SET NOCOUNT ON;
DECLARE @json varchar(max) = (select T.*,
	(select * from PropertyInt WHERE ObjectId=T.ObjectId FOR JSON AUTO) as PropertyInt, 
	(select * from PropertyFloat WHERE ObjectId=T.ObjectId FOR JSON AUTO) as PropertyFloat,
    (select * from PropertyString WHERE ObjectId=T.ObjectId FOR JSON AUTO) as PropertyString, 
	(select * from PropertyBigString WHERE ObjectId=T.ObjectId FOR JSON AUTO) as PropertyBigString
        from {dboName} as T WHERE ObjectId=@ObjectId FOR JSON AUTO);
SET NOCOUNT OFF;
SELECT @json as [Json]";    // trims JSON output to ~ 2K chars
        string json = await dataContext.GetList_ExecDevelopmentScalar<string>(sql, new { o.ObjectId });
        string origin = Request.Query["origin"];
        if (!string.IsNullOrEmpty(origin))
        {
            Response.Headers.Add("Access-Control-Allow-Origin", origin);
        }
        if (string.IsNullOrEmpty(json)) {
            json = "[]";
        }
        return Content(json, "application/json");
    }

    /// <summary>
    /// Get Object Properties as JSON object
    /// </summary>
    /// <param name="objectId">ObjectId - has a priority</param>
    /// <param name="url">ObjectUrl - used if ObjectId==0</param>
    /// <returns>JSON</returns>
    [HttpGet]
    public async Task<IActionResult> GetObjectProperties([FromQuery] int objectId, [FromQuery] string url)
    {
        // https://mdi.matinf.pro/ajax/getobjectproperties?objectid=1
        // https://mdi.matinf.pro/ajax/getobjectproperties?objectid=2
        ObjectInfo o;
        if (objectId != 0)
        {
            o = await dataContext.ObjectInfo_Get(objectId);
        }
        else
        {
            o = await dataContext.GetObjectByUrl(url);
        }
        if (o == null || o.ObjectId == 0)
        {
            return NotFound($"Can not find object ({objectId}, {url})");
        }
        if (HttpContext.IsReadDenied(o.AccessControl, (int)o._createdBy))
        {
            return this.RedirectToLoginOrAccessDenied(HttpContext);
        }
        // string sql = $@"select * from dbo.fn_GetObjectProperties(@ObjectId) order by SortCode FOR JSON AUTO";    // trims JSON output to ~ 2K chars
        string sql = $@"SET NOCOUNT ON;
DECLARE @json varchar(max) = (select * from dbo.fn_GetObjectProperties(@ObjectId) order by SortCode FOR JSON AUTO);
SET NOCOUNT OFF;
SELECT @json as [Json]";
        string json = await dataContext.GetList_ExecDevelopmentScalar<string>(sql, new { o.ObjectId });
        string origin = Request.Query["origin"];
        if (!string.IsNullOrEmpty(origin))
        {
            Response.Headers.Add("Access-Control-Allow-Origin", origin);
        }
        if (string.IsNullOrEmpty(json)) {
            json = "[]";
        }
        return Content(json, "application/json");
    }



    /// <summary>
    /// gets the current state of the request (so far only IsUserAuthentificated for user )
    /// /ajax/getstate
    /// </summary>
    /// <returns></returns>
    [HttpGet]
    public async Task<JsonResult> GetState()
    {
        return new JsonResult(new
        {
            IsUserAuthentificated = Request.Cookies[".AspNetCore.Identity.Application"] != null
        });
    }



    public class ObjectInfoMode : ObjectInfo { 
        public string Mode { get; set; }
    }


    /// <summary>
    /// Gets previous and next object with typeId (Sample) specified
    /// /ajax/GetAdjacentObjectsForUser/?objectId=0&externalId=0&typeId=6
    /// /ajax/GetAdjacentObjectsForUser/?objectId=0&externalId=4450&typeId=6
    /// </summary>
    /// <param name="objectId">current ObjectId to consider (if 0, then currentExternalId must be set)</param>
    /// <param name="externalId">current ExternalId to consider (if ObjectId==0)</param>
    /// <param name="typeId">TypeId filter: 0 - no filter; !=0 - specific object type</param>
    /// <param name="userId">null - get user from HTTP context; 0 - no filter on user; !=0 - objects for specific user</param>
    /// <returns>two rows (or less): with prev and next ObjectInfo</returns>
    [HttpGet]
    public async Task<JsonResult> GetAdjacentObjectsForUser(int objectId, int externalId, int typeId = DataContext.SampleTypeId, int? userId = null)
    {
        ObjectInfoMode? prevObject = null, nextObject = null;
        if (userId == null) {
            userId = HttpContext.GetUserId();
        }
        if (objectId == 0) {
            ObjectInfo? o = await dataContext.ObjectInfo_GetByExternalId(externalId, typeId);
            objectId = o?.ObjectId ?? 0;
        }
        // gets from 0 to 2 rows
        var list = await dataContext.GetList_ExecDevelopment<ObjectInfoMode>(
$@"SELECT * FROM (select TOP 1 'prev' as Mode, * from dbo.ObjectInfo WHERE TenantId=@TenantId {(typeId != 0 ? " AND TypeId=@typeId" : string.Empty)} {(userId != 0 ? " AND _createdBy=@userId" : string.Empty)}
AND ObjectId<@objectId
ORDER BY _created DESC) as p

UNION ALL

SELECT * FROM (select TOP 1 'next' as Mode, * from dbo.ObjectInfo WHERE TenantId=@TenantId {(typeId != 0 ? " AND TypeId=@typeId" : string.Empty)} {(userId!=0 ? " AND _createdBy=@userId" : string.Empty)}
AND ObjectId>@objectId
ORDER BY _created ASC) as n
",
            new { dataContext.TenantId, objectId, typeId, userId });
        prevObject = list.FirstOrDefault(x => x.Mode == "prev", (ObjectInfoMode?)null);
        nextObject = list.FirstOrDefault(x => x.Mode == "next", (ObjectInfoMode?)null);
        return new JsonResult(new
        {
            Previous = prevObject, Next = nextObject
        });
    }


    /// <summary>
    /// extend graph by adding nodes and links
    /// </summary>
    /// <param name="objectId">core object (for which we are looking for parents and children)</param>
    /// <returns>JSON</returns>
    [HttpGet]
    public async Task<ActionResult> GetAdjacentNodesAndLinks(int objectId) {
        // int userId = HttpContext.GetUserId();
        UserContext userCtx = UserUtils.GetUserContext(HttpContext);
        AccessControlFilter accessControlFilter = new AccessControlFilter(userCtx.access, userCtx.userId);

        HashSet<GraphNode> lNodes = new HashSet<GraphNode>();   // all nodes in the graph

        // current node (already there)
        // lNodes.Add(new GraphNode(id: $"o{obj.ObjectId}", typeId: obj.TypeId, bucket: "CORE", name: obj.ObjectName));

        // parents: o<ObjectId>
        List<ObjectInfoLinked> listParent = await dataContext.ObjectLinkObject_GetLinkedObjects_Reverse(objectId, accessControlFilter);
        bool showMultipleLinkTypesParent = listParent.Select(x => x.LinkTypeObjectId ?? 0).Distinct().Count() > 1;
        listParent.ForEach(x => {
            lNodes.Add(new GraphNode(id: $"o{x.ObjectId}", typeId: x.TypeId, bucket: "parent", name: x.ObjectName));
        });

        // children: o<ObjectId>
        List<ObjectInfoLinked> listChildren = await dataContext.ObjectLinkObject_GetLinkedObjects(objectId, accessControlFilter);
        bool showMultipleLinkTypesChildren = listChildren.Select(x => x.LinkTypeObjectId ?? 0).Distinct().Count() > 1;
        listChildren.ForEach(x => {
            lNodes.Add(new GraphNode(id: $"o{x.ObjectId}", typeId: x.TypeId, bucket: "children", name: x.ObjectName));
        });

        // grouping nodes:
        var groups = lNodes.GroupBy(x => $"{x.Bucket}_{x.TypeId}");
        foreach (IGrouping<string, GraphNode> group in groups)  // iterate through groups
        {
            // group.Key
            int count = group.Count();
            if (count > 1)
            {
                string id = $"g_{group.Key}";
                foreach (var item in lNodes.Where(x => group.Key == $"{x.Bucket}_{x.TypeId}"))
                {
                    item.ParentId = id;
                }
                lNodes.Add(new GraphNode(id: id, typeId: 0, bucket: "cluster", name: string.Empty));
            }
            // foreach (var item in group) { }
        }

        // Summary
        bool showMultipleLinkTypes = showMultipleLinkTypesParent || showMultipleLinkTypesChildren;
        // public override string ToString() => $"Link for {base.ToString()} [ObjectLinkObjectId={ObjectLinkObjectId}, MainObjectId={MainObjectId}, LinkTypeObjectId={LinkTypeObjectId}, LinkType={LinkTypeObjectName}]";

        // Nodes == Vertices (form nodes from set of GraphNode elements): o<ObjectId>
        string vertices = string.Join($",{Environment.NewLine}",
            //lNodes.Where(x => !x.Id.Contains("_")).Select(x => $"{{data: {{ id: '{x.Id}', label: '{x.Name}' }}}},")
            lNodes.Select(x => x.ToJsonString()/*$"{{data: {{ id: '{x.Id}', label: '{x.Name}' }}}}"*/)
        );

        // Links == Edges (form links from listParent+listChildren): olo<ObjectLinkObjectId>_o<ObjectId1>_o<ObjectId2>
        string edges = string.Join($",{Environment.NewLine}",
            Enumerable.Union(
                listParent.Select(x => x.ToJsonParentString(objectId) /*$"{{data: {{ id: 'olo{x.ObjectLinkObjectId}_o{x.ObjectId}_o{obj.ObjectId}', source: 'o{x.ObjectId}', target: 'o{obj.ObjectId}', label: '{x.LinkTypeObjectName}' }}}}"*/)
                ,
                listChildren.Select(x => x.ToJsonChildString(objectId) /*$"{{data: {{ id: 'olo{x.ObjectLinkObjectId}_o{obj.ObjectId}_o{x.ObjectId}', source: 'o{obj.ObjectId}', target: 'o{x.ObjectId}', label: '{x.LinkTypeObjectName}' }}}}"*/)
            )
        );
        string json = $@"[
{vertices}
,
{edges}
]";
        return new ContentResult()
        {
            Content = json,//.Replace(Environment.NewLine, string.Empty),
            ContentType = "application/json"
        };
    }

}
