﻿using System.Data;
using Dapper;
using InfProject.Models;
using WebUtilsLib;
using Microsoft.Data.SqlClient;

namespace InfProject.DBContext;

public partial class DataContext
{
    /// <summary>
    /// Scans DB for available chemical elements, regarding currently selected filter (chemical system, SearchPhrase, etc)
    /// </summary>
    /// <param name="filter">Filter</param>
    /// <returns>list of available for choosing chemical elements</returns>
    public async Task<List<string>> GetAvailableElements(Filter filter)
    {
        return await SearchByFilter<string>(SearchStoredProcedure_Filter.GetElements_Filter, filter);
    }

    public async Task<List<T>> SearchByFilter<T>(SearchStoredProcedure_Filter spMode, Filter? filter = null)
    {
        //if (filter == null || filter.IsEmpty)
        //    return await GetSamples();

        using (IDbConnection connection = new SqlConnection(ConnectionString))
        {
            var FilterCompositionItems = filter.CompositionItems.ToDataTable().AsTableValuedParameter("dbo.FilterCompositionItem");
            var FilterPropertyItems = filter.PropertyItems.ToDataTable().AsTableValuedParameter("dbo.FilterPropertyItem");
            IEnumerable<T> res = await connection.QueryAsync<T>($@"EXEC dbo.{spMode} @TenantId, @AccessControl, @UserId, @FilterCompositionItems, @FilterPropertyItems, @TypeId, @SearchPhrase, @CreatedByUser, @CreatedMin, @CreatedMax, @ObjectId, @ExternalId, @JSON",
            new
            {
                TenantId,
                filter.AccessFilter.AccessControl,
                filter.AccessFilter.UserId,
                FilterCompositionItems,
                FilterPropertyItems,
                filter.TypeId,
                filter.SearchPhrase,
                filter.CreatedByUser,
                filter.CreatedMin,
                filter.CreatedMax,
                filter.ObjectId,
                filter.ExternalId,
                JSON = filter.Json
            });
            return res.ToList();
        }
    }

    /// <summary>
    /// delete Compositions for EDX file (after generalization - any nested objects, filter by TypeId is possbile)
    /// </summary>
    /// <param name="edxObjectId">ObjectId of EDX</param>
    /// <param name="compositionTypeId">TypeId for nested (children) Composition objects</param>
    /// <returns></returns>
    public async Task<int> EdxCompositions_Delete(int edxObjectId, int compositionTypeId=0)
    {
        using (IDbConnection connection = new SqlConnection(ConnectionString))
        {
            int retVal = await connection.ExecuteAsync("EXEC dbo.Delete_NestedObjects @TenantId, @ObjectId, @TypeId",
            new { TenantId, ObjectId=edxObjectId, TypeId=compositionTypeId });
            return retVal;
        }
    }

    /// <summary>
    /// ObjectId==SampleId for Measurement Area
    /// </summary>
    /// <param name="typeId_MeasurementArea">"Measurement Area" or "Composition" TypeId</param>
    /// <param name="edxObjectId">objectId for file with EDX data (EDX CSV type)</param>
    /// <param name="measurementArea">value of Integer Property called "Measurement Area"</param>
    /// <returns>ObjectId==SampleId (if != 0 - object found; ==0 - not found)</returns>
    public async Task<int> FindEdxCompositionByMeasurementArea(int typeId_MeasurementArea, int edxCsvObjectId, int measurementArea)
    {
        const string measurementAreaPropertyName = "Measurement Area";
        using (IDbConnection connection = new SqlConnection(ConnectionString))
        {
            int retVal = await connection.ExecuteScalarAsync<int>("SELECT dbo.fn_GetMeasurementAreaObjectId(@TenantId, @TypeId_MeasurementArea, @edxCsvObjectId, @measurementAreaPropertyName, @measurementArea)",
            new { TenantId, typeId_MeasurementArea, edxCsvObjectId, measurementAreaPropertyName, measurementArea });
            return retVal;
        }
    }
}
