﻿using WebUtilsLib;
using TypeValidationLibrary;
using Newtonsoft.Json.Linq;
using InfProject.DTO;
using InfProject.DBContext;

namespace InfProject.Models;

/// <summary>
/// Extensions for TypeInfo class & Tenant
/// </summary>
public static class SettingsExtension
{
    /// <summary>
    /// Gets settings from JSON document
    /// </summary>
    /// <param name="settingsJson">settings in JSON as a string</param>
    /// <returns>JObject</returns>
    public static JObject GetJsonSettings(string? settingsJson) {
        if (string.IsNullOrEmpty(settingsJson))
            return new JObject();
        var json = JObject.Parse(settingsJson);
        return json;
    }

    /// <summary>
    /// Gets settings from JSON document from Type information
    /// </summary>
    /// <param name="type">Object Type</param>
    /// <returns>JObject</returns>
    public static JObject GetJsonSettings(this TypeInfo type) => GetJsonSettings(type.SettingsJson);

    /// <summary>
    /// Gets settings from JSON document from Tenant information
    /// </summary>
    /// <param name="tenant"></param>
    /// <returns></returns>
    public static JObject GetJsonSettings(this Tenant tenant) => GetJsonSettings(tenant.SettingsJson);


    /// <summary>
    /// Gets string settings from JSON document
    /// </summary>
    /// <param name="settingsJson">settings in JSON as a string</param>
    /// <param name="name">parameter name from json</param>
    /// <returns>parameter value or string.empty</returns>
    public static string GetJsonSettingsString(string settingsJson, string name)
    {
        var json = GetJsonSettings(settingsJson);
        return json[name]?.ToString() ?? string.Empty;
    }


    /// <summary>
    /// Gets integer settings from JSON document
    /// </summary>
    /// <param name="settingsJson">settings in JSON as a string</param>
    /// <param name="name">parameter name from json</param>
    /// <returns>parameter value or 0</returns>
    public static int GetJsonSettingsInteger(string settingsJson, string name)
    {
        var valueString = GetJsonSettingsString(settingsJson, name);
        int.TryParse(valueString, out var value);
        return value;
    }

    /// <summary>
    /// Gets string settings from JSON document from Type information
    /// </summary>
    /// <param name="type">Object Type</param>
    /// <param name="name">parameter name from json</param>
    /// <returns>parameter value or string.empty</returns>
    public static string GetJsonSettingsString(this TypeInfo type, string name) => GetJsonSettingsString(type.SettingsJson, name);

    /// <summary>
    /// Gets integer settings from JSON document from Type information
    /// </summary>
    /// <param name="type">Object Type</param>
    /// <param name="name">parameter name from json</param>
    /// <returns>parameter value or 0</returns>
    public static int GetJsonSettingsInteger(this TypeInfo type, string name) => GetJsonSettingsInteger(type.SettingsJson, name);


    /// <summary>
    /// Gets string settings from JSON document from Tenant information
    /// </summary>
    /// <param name="tenant">tenant</param>
    /// <param name="name">parameter name from json</param>
    /// <returns>parameter value or string.empty</returns>
    public static string GetJsonSettingsString(this Tenant tenant, string name) => GetJsonSettingsString(tenant.SettingsJson, name);

    /// <summary>
    /// Gets integer settings from JSON document from Tenant information
    /// </summary>
    /// <param name="tenant">tenant</param>
    /// <param name="name">parameter name from json</param>
    /// <returns>parameter value or 0</returns>
    public static int GetJsonSettingsInteger(this Tenant tenant, string name) => GetJsonSettingsInteger(tenant.SettingsJson, name);


    /// <summary>
    /// Gets ApplyForTypeIds array
    /// </summary>
    /// <param name="type">Object Type</param>
    /// <returns>list of ints (type identifiers)</returns>
    public static List<int> GetSettingsApplyForTypeIds(this TypeInfo type) {
        var json = GetJsonSettings(type);
        var value = json["ApplyForTypeIds"];
        var arr = value?.Values().Select(x => int.Parse(x.ToString())).ToList() ?? new List<int>(new int[] { DataContext.SampleTypeId });
        return arr;
    }

    /// <summary>
    /// Gets AllowedExtensions array
    /// </summary>
    /// <param name="type">Object Type</param>
    /// <returns>string with URL for external POST request to visualize the object</returns>
    public static List<string> GetSettingsAllowedExtensions(this TypeInfo type)
    {
        var json = GetJsonSettings(type);
        var value = json["AllowedExtensions"];
        var arr = value?.Values().Select(x => x.ToString()).ToList() ?? new List<string>();
        return arr;
    }

    /// <summary>
    /// Gets URL to visualize the object (or string.Empty). Supposed to be user for VALID files (IFileValidator should return success)
    /// </summary>
    /// <param name="type">Object Type</param>
    /// <returns>string with URL for external POST request to visualize the object</returns>
    public static string GetSettingsUrlPostVisualizer(this TypeInfo type) =>
        GetJsonSettingsString(type, "UrlPostVisualizer");

    /// <summary>
    /// Gets CustomEditPath to provide customized edit experience for a type
    /// </summary>
    /// <param name="type">Object Type</param>
    /// <returns>string with path prefix to the edit form to create/edit the object if a given type</returns>
    public static string GetSettingsCustomEditPath(this TypeInfo type) =>
        GetJsonSettingsString(type, "CustomEditPath");

    /// <summary>
    /// Gets IncludePropertiesFrom settings
    /// </summary>
    /// <param name="type">Object Type</param>
    /// <returns>true - show properties form in object creation/modification UI and update properties accordingly; false - default behavior (without properties form)</returns>
    public static bool GetSettingsIncludePropertiesForm(this TypeInfo type)
    {
        var value = GetJsonSettingsString(type, "IncludePropertiesForm");
        int.TryParse(value, out int retVal);
        return retVal!=0;
    }

    /// <summary>
    /// Gets ValidationContext settings
    /// </summary>
    /// <param name="type">Object Type</param>
    /// <returns>null or String.Empty == no context; "System" - include chemical system in "ContextSystem" header</returns>
    public static string GetSettingsValidationContext(this TypeInfo type)
    {
        var value = GetJsonSettingsString(type, "ValidationContext");
        return value;
    }
}
