﻿using WebUtilsLib;
using TypeCoreLibrary;
using Newtonsoft.Json.Linq;
using Microsoft.AspNetCore.SignalR;
using static Microsoft.EntityFrameworkCore.DbLoggerCategory;

namespace InfProject.Models;

/// <summary>
/// Extensions for TypeInfo class
/// </summary>
public static class TypeInfoExtension
{
    /// <summary>
    /// Creates an object for validating documents for type
    /// </summary>
    /// <param name="type">Object Type</param>
    /// <param name="currentContextObject">ObjectInfo for ehich we are going to extract values, i.e. the context is determined by it</param>
    /// <returns>IFileValidator</returns>
    public static IFileValidator GetValidator(this TypeInfo type, Context? currentContext)
    {
        IFileValidator obj = null!;
        var schema = type.ValidationSchema;
        if (string.IsNullOrEmpty(schema))
        {  // ok (by default)
            obj = new TypeValidator_Ok();
            var allowedExt = type.GetSettingsAllowedExtensions();
            if (allowedExt?.Count > 0) {
                obj.AllowedExtensions = allowedExt;
            }
            return obj;
        }
        if (schema.StartsWith("type:"))
        { // type:TypeValidationLibrary.TypeValidator_EDX_CSV
            var className = schema.Substring(5);
            obj = ValidatorTypeHelper.CreateObject<IFileValidator>(className);
            var allowedExt = type.GetSettingsAllowedExtensions();
            if (allowedExt?.Count>0) {
                obj.AllowedExtensions = allowedExt;
            }
            EmbedContext(obj, currentContext);
        }
        else if (schema.StartsWith("http://") || schema.StartsWith("https://"))
        {
            obj = new HttpTypeValidatorProxy(schema);
            // TODO: EmbedContext HERE!!!
        }
        return obj;
    }


    /// <summary>
    /// Creates an IFileValidator object for validation from schema string
    /// </summary>
    /// <param name="schema">schema (starts with "type:" or "http(s)://")</param>
    /// <returns>IFileValidator</returns>
    public static IFileValidator GetValidator(string? schema)
    {
        IFileValidator obj = null!;
        if (string.IsNullOrEmpty(schema))
        {  // ok (by default)
            obj = new TypeValidator_Ok();
            return obj;
        }
        if (schema.StartsWith("type:"))
        {
            var className = schema.Substring(5);
            obj = ValidatorTypeHelper.CreateObject<IFileValidator>(className);
        }
        else if (schema.StartsWith("http://") || schema.StartsWith("https://"))
        {
            obj = new HttpTypeValidatorProxy(schema);
        }
        return obj;
    }



    /// <summary>
    /// Creates an object for getting full table data
    /// </summary>
    /// <param name="type">Object Type</param>
    /// <param name="validator"></param>
    /// <returns>ITableGetter</returns>
    public static ITableGetter? GetTableGetter(this TypeInfo type, IFileValidator validator = null)
    {
        ITableGetter obj = validator as ITableGetter;
        if (obj != null)
        {
            return obj;
        }
        var schema = type.DataSchema;
        if (string.IsNullOrEmpty(schema))
        {  
            return null;
        }
        if (schema.StartsWith("type:"))
        { // type:TypeValidationLibrary.TypeValidator_EDX_CSV
            var className = schema.Substring(5);
            obj = ValidatorTypeHelper.CreateObject<ITableGetter>(className);
        }
        else if (schema.StartsWith("http://") || schema.StartsWith("https://"))
        {
            obj = new HttpTypeDataProxy(schema);
        }
        return obj;
    }


    /// <summary>
    /// Creates an object for getting full table data
    /// </summary>
    /// <param name="type">Object Type</param>
    /// <param name="prevValidatorObject">validator object, that can be also a data extractor object (no repeatable objects creation => performance boosting, )</param>
    /// <param name="currentContextObject">ObjectInfo for ehich we are going to extract values, i.e. the context is determined by it</param>
    /// <returns>IDatabaseValuesGetter?</returns>
    public static IDatabaseValuesGetter? GetDatabaseValuesGetter(this TypeInfo type, object prevValidatorObject, Context? currentContext)
    {
        EmbedContext(prevValidatorObject as IContext, currentContext);
        IDatabaseValuesGetter obj = prevValidatorObject as IDatabaseValuesGetter;
        if (obj != null)
        {
            return obj;
        }
        var schema = type.DataSchema;
        if (string.IsNullOrEmpty(schema))   // https://validation.matinf.pro/edx/data
        {
            return null;
        }
        if (schema.StartsWith("type:"))
        { // type:TypeValidationLibrary.TypeValidator_EDX_CSV
            var className = schema.Substring(5);
            obj = ValidatorTypeHelper.CreateObject<IDatabaseValuesGetter>(className);
        }
        else if (schema.StartsWith("http://") || schema.StartsWith("https://"))
        {
            obj = new HttpTypeDataProxy(schema);
        }
        return obj;
    }


    /// <summary>
    /// Gets Context by calling contextPrecursor.GetContextAsync if contextPrecursor.ValidationContext is not empty
    /// </summary>
    /// <param name="contextPrecursor">ContextPrecursor</param>
    /// <returns>Context or null</returns>
    private static TypeCoreLibrary.Context? GetContextImplementation(ContextPrecursor? contextPrecursor = null) {
        TypeCoreLibrary.Context? context = null;
        // TODO: Get DB Connection & extract Context
        if (!string.IsNullOrEmpty(contextPrecursor?.ValidationContext) && contextPrecursor?.GetContextAsync != null) {
            context = contextPrecursor.GetContextAsync(contextPrecursor).GetAwaiter().GetResult();
        }
        return context;
    }

    /// <summary>
    /// Get Context according to the type settings in JSON.ValidationContext and ContextPrecursor 
    /// </summary>
    /// <param name="type">TypeInfo</param>
    /// <param name="contextPrecursor">ContextPrecursor</param>
    /// <returns>Context</returns>
    private static TypeCoreLibrary.Context? GetContext(this TypeInfo type, ContextPrecursor? contextPrecursor = null) {
        // TypeInfo.SettingsJson='{ "AllowedExtensions": [".txt",".csv"], "ValidationContext": "ChemicalSystem" }'
        string validationContext = type.GetSettingsValidationContext();
        if (contextPrecursor != null)
            contextPrecursor.ValidationContext = validationContext;
        TypeCoreLibrary.Context? context = GetContextImplementation(contextPrecursor);
        return context;
    }


    /// <summary>
    /// Validates file of type by given absolute file path
    /// </summary>
    /// <param name="type">Object Type</param>
    /// <param name="filePath">absolute file path</param>
    /// <param name="contextPrecursor">identifiers from DB to get the context, if neccesary</param>
    /// <returns>TypeValidatorResult</returns>
    public static TypeValidatorResult ValidateFile(this TypeInfo type, string filePath, ContextPrecursor? contextPrecursor = null)
    {
        if (string.IsNullOrEmpty(filePath)) {
            if (type.FileRequired) {
                return new TypeValidatorResult(404, $"FileRequired is set for type {type.TypeName}, but no file was provided");
            }
            return new TypeValidatorResult();   // success
        }

        TypeCoreLibrary.Context? context = type.GetContext(contextPrecursor);
        var validator = type.GetValidator(context);
        validator.File = new FileInfo(filePath);
        var res = validator.Validate();
        return res;
    }
    /// <summary>
    /// Validates object data of type by given stream
    /// </summary>
    /// <param name="type">Object Type</param>
    /// <param name="inputStream">stream with object data</param>
    /// <returns></returns>
    public static TypeValidatorResult ValidateStream(this TypeInfo type, Stream inputStream)
    {
        if (inputStream==null)
        {
            if (type.FileRequired) {
                return new TypeValidatorResult(404, $"FileRequired is set for type {type.TypeName}, but no file was provided");
            }
            return new TypeValidatorResult();
        }
        var validator = type.GetValidator(null);    // TODO: implement context: null => ObjectInfo 
        validator.Stream = inputStream;
        var res = validator.Validate();
        return res;
    }

    /// <summary>
    /// Validates file and (if ok) returns DatabaseValues
    /// </summary>
    /// <param name="type">Object Type</param>
    /// <param name="absoluteFilePath">absolute path to the file</param>
    /// <param name="contextPrecursor">ContextPrecursor used to extract context if required</param>
    /// <returns></returns>
    public static (TypeValidatorResult res, DBUtils.DatabaseValues data) ValidateFileAndGetDataValues(this TypeInfo type, string absoluteFilePath, ContextPrecursor? contextPrecursor = null)
    {
        TypeValidatorResult res;
        DBUtils.DatabaseValues data;
        if (string.IsNullOrEmpty(absoluteFilePath))
        {
            if (type.FileRequired)
            {
                res = new TypeValidatorResult() { Code = 404, Message = $"required file is not defined [ObjectFilePath is empty]" };
            }
            else {
                res = new TypeValidatorResult(); // ok
            }
            data = new DBUtils.DatabaseValues() { DeletePreviousProperties = false, DeletePreviousCompositions = false }; // important not to delete all data if no file is there

        }
        else {
            TypeCoreLibrary.Context? currentContext = type.GetContext(contextPrecursor);
            var validator = type.GetValidator(currentContext);
            validator.File = new FileInfo(absoluteFilePath);
            res = validator.Validate();
            if (!res) {
                return (res, new DBUtils.DatabaseValues() { DeletePreviousProperties = false, DeletePreviousCompositions = false });
            }
            var dbValuesGetter = type.GetDatabaseValuesGetter(validator, currentContext);   // could be other object
            if (dbValuesGetter != null)
            {
                dbValuesGetter.File = new FileInfo(absoluteFilePath);   // in case it is other object
                data = dbValuesGetter.GetDatabaseValues();
            }
            else
            {
                data = new DBUtils.DatabaseValues();
            }
        }
        return (res, data);
    }


    /// <summary>
    /// Set Up Context (important to correctly validate or import data
    /// </summary>
    /// <param name="extractorContext">IContext interface of data extractor</param>
    /// <param name="currentContext"></param>
    private static void EmbedContext(IContext? extractorContext, Context? currentContext)
    {
        if (extractorContext == null || currentContext == null)
            return;
        extractorContext.Context = currentContext;

        // //TODO: Set Context
        // 0. Extract currentContextObject in Dictionary
        //Dictionary<string, string> dict = ExtractContextToDictionary(currentContextObject);
        // 1. Fill Dictionary 
        // 2. Call SetContext
    }


    //public static Dictionary<string, string> ExtractContextToDictionary(this DataContext dataContext, ObjectInfo currentContextObject) {
    //    Dictionary<string, string> dict = new Dictionary<string, string>();


    //    return dict;
    //}

}
