﻿using Azure.Core;
using Azure.Core.GeoJson;
using InfProject.DBContext;
using InfProject.Models;
using Microsoft.AspNetCore.Identity.UI.Services;
using System.Data;
using System.Net.Mail;

namespace InfProject.Utils
{
    public class MailHelper
    {
        private SimpleMailSender mailSender;
        private HttpContext httpContext;
        private DataContext dataContext;
        public MailHelper(SimpleMailSender mailSender, HttpContext httpContext, DataContext dataContext) {
            this.mailSender = mailSender;
            this.httpContext = httpContext;
            this.dataContext = dataContext;
        }

        /// <summary>
        /// Host name only
        /// </summary>
        public string Host => httpContext.Request.Host.ToString();

        /// <summary>
        /// Tenant URL
        /// </summary>
        public string URL => $"http{(httpContext.Request.IsHttps ? "s" : string.Empty)}://{Host}";

        public string Footer => $"<br /><br /><a href=\"{URL}\">{Host}</a> automatic notification service";

        /// <summary>
        /// send e-Mail regarding handover event
        /// </summary>
        /// <param name="handover">handover event object</param>
        /// <param name="sample">sample</param>
        /// <param name="senderUser">sender</param>
        /// <param name="destinationUser">target user</param>
        /// <returns></returns>
        public async Task SendHandoverMail(Handover handover, ObjectInfo sample, WebAppUser senderUser, WebAppUser destinationUser) {
            string htmlMessage;
            if (handover.DestinationConfirmed.HasValue) {   // STEP 2: confirmation (on sample receivement)
                htmlMessage = $@"<p>Dear {senderUser.Name},<br /><br />
<a href=""mailto:{destinationUser.Email}"">{destinationUser.Name}{(string.IsNullOrEmpty(destinationUser.Project) ? string.Empty : $" {destinationUser.Project}")}</a> confirmed your handover receivement: {handover.ObjectName}<br />
It was done on {handover.DestinationConfirmed} with the folliwing comment: <i>{handover.DestinationComments}</i>.
{Footer}
</p>";
                // send letter to Destination User;
                await mailSender.SendEmailAsync(senderUser.Email, subject: $"Handover confirmed: {handover.ObjectName}", htmlMessage: htmlMessage);
            }
            else { // STEP 1: initial event spawned
                htmlMessage = $@"<p>Dear {destinationUser.Name},<br /><br />
<a href=""mailto:{senderUser.Email}"">{senderUser.Name}{(string.IsNullOrEmpty(senderUser.Project) ? string.Empty : $" {senderUser.Project}")}</a> registered a handover event for you: {handover.ObjectName}<br />
{(handover.Amount>0 ? $"Amount: <i>{handover.Amount} {handover.MeasurementUnit}</i>.<br />" : string.Empty)}
Comment: <i>{(string.IsNullOrEmpty(handover.ObjectDescription) ? "not specified" : handover.ObjectDescription)}</i>.<br />
When you receive the sample, please <a href=""{URL}/object/{sample.ObjectNameUrl}"">confirm this</a> in the system by specifying your comments for sender!
{Footer}
</p>";
                // send letter to Destination User;
                await mailSender.SendEmailAsync(destinationUser.Email, subject: $"Handover event registered: {handover.ObjectName}", htmlMessage: htmlMessage);
            }
        }



        /// <summary>
        /// send a notification e-Mail regarding an object update/creation
        /// </summary>
        /// <param name="type">Object type</param>
        /// <param name="@object">@object</param>
        /// <param name="userCreator">sender</param>
        /// <param name="userForNotification">target user</param>
        /// <param name="cc">list od e-mails for "cc" (copy)</param>
        /// <returns></returns>
        public async Task SendNotificationMail(TypeInfo type, ObjectInfo @object, WebAppUser userCreator, WebAppUser userForNotification, string[] cc = null, Attachment[] attachments = null)
        {
            string htmlMessage = $@"<p>Dear {userForNotification.Name},<br /><br />
<a href=""mailto:{userCreator.Email}"">{userCreator.Name}{(string.IsNullOrEmpty(userCreator.Project) ? string.Empty : $" {userCreator.Project}")}</a> created an object [{type.TypeName}]: <a href=""{URL}/object/{@object.ObjectNameUrl}"">{@object.ObjectName}</a><br />
{(string.IsNullOrEmpty(@object.ObjectDescription) ? "Description not specified" : $"Description: <i>{@object.ObjectDescription}</i>" )}.<br />
{Footer}
</p>";
            // send letter to userForNotification User;
            await mailSender.SendEmailAsync(userForNotification.Email, subject: $"{type.TypeName} created: {@object.ObjectName}", htmlMessage: htmlMessage, cc, attachments);
        }


        /// <summary>
        /// E-mail notification on object creation
        /// </summary>
        /// <param name="creatorUserId"></param>
        /// <param name="notificationUserId"></param>
        /// <param name="notificationProject"></param>
        /// <param name="dbType"></param>
        /// <param name="object"></param>
        /// <returns></returns>
        public async Task SendNotificationMailWrapper(int creatorUserId, int notificationUserId, string notificationProject, TypeInfo dbType, ObjectInfo @object) {
            string EmailNotificationSetting = dbType.GetJsonSettings().SelectToken("OnObjectCreated.EmailNotification")?.ToString() ?? string.Empty;
            if (notificationUserId == 0 && EmailNotificationSetting == "user" || string.IsNullOrEmpty(notificationProject) && EmailNotificationSetting == "project")
                return;
            WebAppUser userCreator = await dataContext.GetUser(creatorUserId);  // object creator
            List<WebAppUser> users;
            if ((EmailNotificationSetting == "user" || EmailNotificationSetting == "userAndPI") && notificationUserId != 0) {
                WebAppUser notificationUser = await dataContext.GetUser(notificationUserId);
                string[] cc = null;
                if (EmailNotificationSetting == "user" && !string.IsNullOrEmpty(notificationUser.Email))
                {   // send a user-only notification
                    await SendNotificationMail(dbType, @object, userCreator, notificationUser);
                } 
                else if (EmailNotificationSetting == "userAndPI" && !string.IsNullOrEmpty(notificationUser.Email))
                {   // send a notification to a user + PI in cc
                    IEnumerable<int> principalInvestigators = await dataContext.GetUserIdsByClaim(claimType: "PI", claimValue: "1");
                    string[] Projects = notificationUser.Projects;
                    users = await dataContext.GetUserListAll();
                    HashSet<string> ccEmail = new HashSet<string>();
                    foreach (string project in notificationUser.Projects)   // iterate through projects of the user
                    {
                        // get PIs for a project
                        var activeProjectUsersAndPI = users.Where(x => x.Projects.Contains(project) && principalInvestigators.Contains(x.Id) && !string.IsNullOrEmpty(x.Email) && (!x.LockoutEnabled || !x.LockoutEnd.HasValue || x.LockoutEnd.Value > DateTime.SpecifyKind(DateTime.Now, DateTimeKind.Local))).ToList();
                        foreach (var pi in activeProjectUsersAndPI)
                        {
                            if (!string.IsNullOrEmpty(pi.Email)) {
                                ccEmail.Add(pi.Email);
                            }
                        }
                        cc = ccEmail.ToArray();
                    }
                    //var activeProjectUsers = users.Where(x => x.Projects.Contains(notificationProject) && !string.IsNullOrEmpty(x.Email) && (!x.LockoutEnabled || !x.LockoutEnd.HasValue || x.LockoutEnd.Value > DateTime.SpecifyKind(DateTime.Now, DateTimeKind.Local))).ToList();
                    await SendNotificationMail(dbType, @object, userCreator, notificationUser, cc, attachments: null);
                }
            }
            else if (EmailNotificationSetting == "project" && !string.IsNullOrEmpty(notificationProject))
            {
                users = await dataContext.GetUserListAll();
                var activeProjectUsers = users.Where(x => x.Projects.Contains(notificationProject) && !string.IsNullOrEmpty(x.Email) && (!x.LockoutEnabled || !x.LockoutEnd.HasValue || x.LockoutEnd.Value > DateTime.SpecifyKind(DateTime.Now, DateTimeKind.Local))).ToList();
                for (int i = 0; i < activeProjectUsers.Count; i++)
                {   // send notification
                    await SendNotificationMail(dbType, @object, userCreator, activeProjectUsers[i]);
                }
            }
        }


    }
}
