﻿using System.Collections.Generic;
using System.Data;
using System.IO;
using System.Linq;
using WebUtilsLib;
using TypeCoreLibrary;
using static TypeValidationLibrary.TypeValidator_EDX_CSV;
using static WebUtilsLib.DBUtils;
using Newtonsoft.Json.Linq;
using System.Xml.Linq;
using UglyToad.PdfPig;
using Microsoft.AspNetCore.Components.Forms;
using System.Globalization;
using System.Text.RegularExpressions;

// TypeInfo.ValidationSchema = type:TypeValidationLibrary.TypeValidator_EDX_CSV
namespace TypeValidationLibrary;

/// <summary>
/// base class
/// </summary>
public class TypeValidator_BET : TypeValidatorBaseMdi, IFileValidator, ITableGetter, IDatabaseValuesGetter
{


    private static List<(string name, double value, string comment)>? ExtractValues_BET(string fileName, Stream inputData, string? fileNameFromHeader=null) {
        static string ExtractValue(string source, string startPattern, string endPattern)
        {
            if (string.IsNullOrEmpty(source) || string.IsNullOrEmpty(startPattern) || string.IsNullOrEmpty(endPattern))
            {
                return string.Empty;
            }

            int startIndex = source.IndexOf(startPattern);
            if (startIndex == -1)
            {
                return string.Empty; // no startPattern
            }

            startIndex += startPattern.Length; // move index after startPattern

            int endIndex = source.IndexOf(endPattern, startIndex);
            if (endIndex == -1)
            {
                return string.Empty; // no endPattern
            }

            return source.Substring(startIndex, endIndex - startIndex);
        }

        static (double r, double area) ExtractValues_BET_Body(string fileName, Stream inputData, string? fileNameFromHeader = null)
        {
            double r = 0, area = 0;
            if ((fileName?.EndsWith(".pdf") ?? false) || (fileNameFromHeader?.EndsWith(".pdf") ?? false))
            {
                (r, area) = ExtractValues_BET_Body_Pdf(fileName, inputData, fileNameFromHeader);
            }
            else {
                (r, area) = ExtractValues_BET_Body_Txt(fileName, inputData, fileNameFromHeader);
            }
            return (r, area);
        }


        static (double r, double area) ExtractValues_BET_Body_Pdf(string fileName, Stream inputData, string? fileNameFromHeader = null)
        {
            double r = 0, area = 0;
            using (PdfDocument pdf = string.IsNullOrEmpty(fileName) ? PdfDocument.Open(inputData) : PdfDocument.Open(fileName))
            {
                //foreach (var page in pdf.GetPages())
                //{
                //    Console.WriteLine($"=== page{page.Number} ===");
                //    string text = page.Text;
                //    Console.WriteLine(text);
                //}
                var page = pdf.GetPages().FirstOrDefault();
                string rSt = null, areaSt = null;
                string text = page.Text;

                if (page.Number != 1)
                {
                    throw new Exception("page number must be 1");
                }
                if (!text.StartsWith("Anton Paar Kaomi for Nova"))
                {
                    throw new Exception("Text on the first page should start with \"Anton Paar Kaomi for Nova\"");
                }
                if (!text.Contains("Multipoint BET Summary/Results"))
                {
                    throw new Exception("Text on the first page should contain \"Multipoint BET Summary/Results\"");
                }

                // Isotherm BranchIsotherm BranchAdsorptionSlopeSlope42.65InterceptIntercept0.25184Correlation coeff., rCorrelation coeff., r0.999978C constantC constant170.353Surface areaSurface area81.174 m?/g
                rSt = ExtractValue(text, "Correlation coeff., rCorrelation coeff., r", "C constantC constant");
                areaSt = ExtractValue(text, "Surface areaSurface area", "m²/g");
                double.TryParse(rSt, out r);
                double.TryParse(areaSt, out area);

                // Console.WriteLine($"r = {r}");
                // Console.WriteLine($"area = {area}");
            }
            return (r, area);
        }


        static (double r, double area) ExtractValues_BET_Body_Txt(string fileName, Stream inputData, string? fileNameFromHeader = null)
        {
            double r = 0, area = 0;
            if (string.IsNullOrEmpty(fileName))
            {
                (r, area) = ParseMbetData(inputData);
            }
            else {
                (r, area) = ParseMbetDataFile(fileName);
            }
            return (r, area);
        }



        static (double CorrelationR, double SurfaceArea) ParseMbetDataFile(string filePath)
        {
            using var reader = new StreamReader(filePath);
            return ParseFromStream(reader);
        }

        static (double CorrelationR, double SurfaceArea) ParseMbetData(Stream stream)
        {
            using var reader = new StreamReader(stream);
            return ParseFromStream(reader);
        }

        static (double CorrelationR, double SurfaceArea) ParseFromStream(StreamReader reader)
        {
            double? correlationR = null;
            double? surfaceArea = null;

            string? line;
            var numberStyle = NumberStyles.Float | NumberStyles.AllowThousands;
            var formatProvider = CultureInfo.InvariantCulture;

            while ((line = reader.ReadLine()) != null)
            {
                if (correlationR == null)
                {
                    var matchR = Regex.Match(line, @"Correlation coefficient, r\s*=\s*([+-]?\d*\.?\d+(?:[eE][+-]?\d+)?)");
                    if (matchR.Success)
                    {
                        correlationR = double.Parse(matchR.Groups[1].Value, numberStyle, formatProvider);
                        continue;
                    }
                }

                if (surfaceArea == null)
                {
                    var matchSA = Regex.Match(line, @"Surface Area\s*=\s*([+-]?\d*\.?\d+(?:[eE][+-]?\d+)?)");
                    if (matchSA.Success)
                    {
                        surfaceArea = double.Parse(matchSA.Groups[1].Value, numberStyle, formatProvider);
                        continue;
                    }
                }

                if (correlationR.HasValue && surfaceArea.HasValue)
                    break;
            }

            if (correlationR == null || surfaceArea == null)
                throw new InvalidDataException("Не удалось найти требуемые параметры в файле.");

            return (correlationR.Value, surfaceArea.Value);
        }


        // Protected BET adsorption report (pdf, txt)
        (double r, double area) = ExtractValues_BET_Body(fileName, inputData, fileNameFromHeader);
        if (r != 0 && area != 0)
        {
            return new List<(string name, double value, string comment)> {
                ( "BET_R", r, "Correlation coefficient"),
                ( "BET_Area", area, "Surface area, m²/g")
            };
        }
        return null;
    }


    static readonly string[] allowedExt = new string[] { 
        ".pdf", // request from Leon Müller (CRC247.C02) 
        ".txt"  // request from Nicolas Cosanne (CRC247.A08)
    };

    private DataTable? dt;
    private List<(string name, double value, string comment)>? prop;
    int verifiedVersion;

    public TypeValidator_BET() : base()
    {
        // default configuration
        AllowedExtensions.AddRange(allowedExt);
    }



    /// <summary>
    /// Initialize validation (resets the state)
    /// </summary>
    protected override void InitValidation()
    {
        dt = null;
        prop = null;
        //Console.WriteLine(File.Name);
    }


    /// <summary>
    /// Type-specific logic
    /// </summary>
    /// <returns></returns>
    protected override TypeValidatorResult ValidateSpecific() {
        InitValidation();
        verifiedVersion = version;
        TypeValidatorResult result = new TypeValidatorResult();

        try
        {
            prop = ExtractValues_BET(File?.FullName, Stream, fileNameFromHeader: FileNameFromHeader);
        }
        catch (Exception ex)
        {
            result.Code = 500;
            result.Message = ex.Message;
        }

        return result;
    }


    /// <summary>
    /// Gets table instance for rendering in a table view
    /// </summary>
    /// <returns>Table</returns>
    public Table GetTable() {
        return null!;
    }

    /// <summary>
    /// get all boundaries values for writing to DB
    /// GetDatabaseValues for IDatabaseValuesGetter
    /// </summary>
    /// <returns></returns>
    public DatabaseValues GetDatabaseValues()
    {
        DatabaseValues dbv = new DatabaseValues();
        if (ValidateSpecific() && prop?.Count>0) {
            GetDatabaseValues_FillProperties(dbv);
        }
        return dbv;
    }


    /// <summary>
    /// get all properties (all values for writing to DB)
    /// </summary>
    /// <param name="dbv">DatabaseValues to fill</param>
    public void GetDatabaseValues_FillProperties(DatabaseValues dbv)
    {
        PropertyValue pv;
        List<PropertyValue> props = dbv.Properties;
        for (int i = 0; i < prop?.Count; i++)
        {
            props.Add(pv = new PropertyValue()
            {
                PropertyType = PropertyType.Float,
                PropertyName = prop[i].name,
                SortCode = 10 + i * 10,
                Row = null,
                Value = prop[i].value,
                Comment = prop[i].comment,
            });
            if (prop[i].name == "BET_Area") {   // add "BET_Area" to the "parent" object(s) (should be a sample)
                dbv.ParentPropertiesAppend.Add(pv);
            }
        }
    }

}