﻿using System.Collections.Generic;
using System.Data;
using System.IO;
using System.Linq;
using WebUtilsLib;
using TypeCoreLibrary;
using static TypeValidationLibrary.TypeValidator_EDX_CSV;
using static WebUtilsLib.DBUtils;
using Newtonsoft.Json.Linq;
using System.Xml.Linq;

// TypeInfo.ValidationSchema = type:TypeValidationLibrary.TypeValidator_EDX_CSV
namespace TypeValidationLibrary;

/*
Index;x;y;Crystal Structure; IDs
1;-22500;-40500;La2O3/Perovskite;"ICSD-7800;ICSD-17668"
2;-18000;-40500;La2O3/Perovskite;"ICSD-7800;ICSD-17668"
3;-13500;-40500;La2O3/Perovskite;"ICSD-7800;ICSD-17668"
4;-9000;-40500;La2O3/Perovskite;"ICSD-7800;ICSD-17668"
5;-4500;-40500;La2O3/Perovskite;"ICSD-7800;ICSD-17668"
6;0;-40500;Perovskite;ICSD-17668
7;4500;-40500;Perovskite;ICSD-17668
8;9000;-40500;Perovskite;ICSD-17668
*/

/// <summary>
/// base class
/// </summary>
public class TypeValidator_XRD_CSV : TypeValidatorBaseMdi, IFileValidator, ITableGetter, IDatabaseValuesGetter
{
    static readonly string[] allowedExt = new string[] { ".csv" };
    static readonly char[] separators = { ';', '/' };

    const string idColumnName = "IDs";
    const string friendlyNameColumnName = "Crystal Structure";

    private DataTable? dt;
    int verifiedVersion;

    public TypeValidator_XRD_CSV() : base()
    {
        // default configuration
        AllowedExtensions.AddRange(allowedExt);
    }



    /// <summary>
    /// Initialize validation (resets the state)
    /// </summary>
    protected override void InitValidation()
    {
        dt = null;
        //Console.WriteLine(File.Name);
    }


    /// <summary>
    /// Type-specific logic
    /// </summary>
    /// <returns></returns>
    protected override TypeValidatorResult ValidateSpecific() {
        InitValidation();
        verifiedVersion = version;
        TypeValidatorResult result = new TypeValidatorResult();

        //if (string.Compare(File.Extension, ".txt", true)==0) { }  // sometimes it could be other extension
        string contents = string.Empty;
        if (File != null)
        {
            contents = System.IO.File.ReadAllText(File.FullName, Encoding);
        }
        else if (Stream != null) {
            using (var reader = new StreamReader(Stream, Encoding))
            {
                contents = reader.ReadToEndAsync().Result;
            }
        }
        if (string.IsNullOrEmpty(contents)) {
            result.Code = 400;
            result.Message = "Empty content";
            return result;
        }

        string csv = contents;

        _ValidateCSV(csv, result);  // CSV must be valid
        if (!result || dt == null) {
            return result;
        }

        _ValidateTable(result); // Table must be valid

        return result;
    }


    /// <summary>
    /// validate CSV
    /// </summary>
    /// <param name="csv"></param>
    /// <returns></returns>
    protected TypeValidatorResult _ValidateCSV(string csv, TypeValidatorResult result) {
        if (result == null) {
            result = new TypeValidatorResult();
        }
        dt = null;
        try
        {
            bool firstLineContainsHeader = true;
            string delimiter = CSVParser.GuessDelimiter(csv);
            dt = CSVParser.VB_Parse(csv, delimiter, firstLineContainsHeader, 
                name => elements.Contains(name) || coordsColumns.Contains(name) ? typeof(float) : typeof(string));
        }
        catch (Exception ex)
        {
            result.Code = 500;
            result.Message = $"{ex.GetType()} {ex.Message}";
        }
        return result;
    }


    /// <summary>
    /// Checks that table contains:
    /// - "Index" column;
    /// - "IDs" column (idColumnName);
    /// - Optional (but recommended): "Crystal Structure" column (friendlyNameColumnName)
    /// </summary>
    /// <param name="result">TypeValidatorResult to adjust during chaining validation</param>
    /// <returns>TypeValidatorResult</returns>
    protected TypeValidatorResult _ValidateTable(TypeValidatorResult result) {
        if (result == null)
        {
            result = new TypeValidatorResult();
        }
        List<string> columnsToRemove = new List<string>();
        try
        {
            if (dt==null)
            {
                result.Code = 549;
                result.Message = $"dt==null";
                return result;
            }
            if (dt.Columns.Count < 2)
            {
                result.Code = 550;
                result.Message = $"dt.Columns.Count < 2 [dt.Columns.Count={dt.Columns.Count}]";
                return result;
            }
            if (!dt.Columns.Contains("Index"))
            {
                result.Code = 560;
                result.Message = $"dt.Columns do not contain Index column";
                return result;
            }
            if (!dt.Columns.Contains(idColumnName))
            {
                result.Code = 570;
                result.Message = $"dt.Columns do not contain {idColumnName} column";
                return result;
            }
            if (!dt.Columns.Contains(friendlyNameColumnName))
            {
                result.Warning = $"[{friendlyNameColumnName}] column not found, please use it for friendly name";
            }
        }
        catch (Exception ex)
        {
            result.Code = 500;
            result.Message = $"{ex.GetType()} {ex.Message}";
        }
        return result;
    }


    /// <summary>
    /// DataTable for ITableGetter
    /// </summary>
    /// <returns></returns>
    private DataTable? DataTable => dt;

    /// <summary>
    /// Is there any coordinates ("X (mm)", "Y (mm)", "x", "y")
    /// GetCoordinates for ITableGetter
    /// </summary>
    /// <returns></returns>
    private CoordinatesResult GetCoordinates(DataTable dt)
    {
        if (dt == null)
            return new CoordinatesResult();
        List<string> coordinates = new List<string>();
        foreach (DataColumn column in dt.Columns)
        {
            if (coordsColumns.Contains(column.ColumnName))
                coordinates.Add(column.ColumnName);
        }
        return new CoordinatesResult(coordinates.ToArray());
    }


    /// <summary>
    /// Is there any coordinates ("X (mm)", "Y (mm)", "x", "y")
    /// GetCoordinates for ITableGetter
    /// </summary>
    /// <returns></returns>
    private DependenciesResult GetDependencies(DataTable dt)
    {
        if (dt == null)
            return new DependenciesResult();
        List<string> xColumns = new List<string>();
        foreach (DataColumn column in dt.Columns)
        {
            if (dependencyColumns.Contains(column.ColumnName, StringComparer.OrdinalIgnoreCase))
                xColumns.Add(column.ColumnName);
        }
        return new DependenciesResult(xColumns.ToArray());
    }

    /// <summary>
    /// Gets table instance for rendering in a table view
    /// </summary>
    /// <returns>Table</returns>
    public Table GetTable() {
        if (verifiedVersion != version)
        {
            Validate();
            verifiedVersion = version;
        }
        return new Table(dt, GetCoordinates(dt), GetDependencies(dt));
    }

    /// <summary>
    /// get all boundaries values for writing to DB
    /// GetDatabaseValues for IDatabaseValuesGetter
    /// </summary>
    /// <returns></returns>
    public DatabaseValues GetDatabaseValues()
    {
        DatabaseValues dbv = new DatabaseValues();
        if (dt == null || dt.Columns.Count < 2)
        {
            return dbv;
        }
        GetDatabaseValues_FillProperties(dbv.Properties);

        GetDatabaseValues_FillPropertiesForCompositions(dbv.CompositionsForSampleUpdate);
        return dbv;
    }


    /// <summary>
    /// get all properties (all phases values for writing to DB)
    /// </summary>
    /// <param name="props">properties to fill</param>
    public void GetDatabaseValues_FillProperties(List<PropertyValue> props)
    {
        HashSet<(string id, string friendlyName)> set = new HashSet<(string id, string friendlyName)>();
        // fill the hashset with all 
        bool hasFriendlyNameColumnName = dt.Columns.Contains(friendlyNameColumnName);
        if (hasFriendlyNameColumnName) {    // idColumnName + friendlyNameColumnName
            for (int i=0; i<dt.Rows.Count; i++) {
                DataRow row = dt.Rows[i];
                if (row[idColumnName] != null && row[idColumnName] != DBNull.Value && row[friendlyNameColumnName] != null && row[friendlyNameColumnName] != DBNull.Value)
                {
                    var idsArr = row[idColumnName].ToString().Split(';', StringSplitOptions.RemoveEmptyEntries);
                    var namesArr = row[friendlyNameColumnName].ToString().Split(separators, StringSplitOptions.RemoveEmptyEntries);
                    if (idsArr.Length != namesArr.Length) {
                        throw new Exception($"Array lengths mismatch: [row={i+1}, {idColumnName}.Length={idsArr.Length}; {friendlyNameColumnName}.Length={namesArr.Length}]");
                    }
                    var tuples = idsArr.Zip(namesArr, (id, name) => ((id, name)));
                    foreach (var item in tuples) {
                        set.Add(item);
                    }
                }
            }
        }
        else {  // idColumnName only
            foreach (DataRow row in dt.Rows) {
                if (row[idColumnName] != null && row[idColumnName] != DBNull.Value)
                {
                    Array.ForEach(row[idColumnName].ToString().Split(';', StringSplitOptions.RemoveEmptyEntries), x => { set.Add((x, string.Empty)); });
                }
            }
        }

        List<(string id, string friendlyName)> list = set.ToList();
        for (int i = 0; i < list.Count; i++)
        {
            // identifier
            props.Add(new PropertyValue() {
                PropertyType = PropertyType.String,
                PropertyName = "ID",
                SortCode = 10,
                Row = i+1,
                Value = list[i].id,
                Comment = null
            });

            // friendlyName
            if (!string.IsNullOrEmpty(list[i].friendlyName))
                continue;
            props.Add(new PropertyValue() {
                PropertyType = PropertyType.String,
                PropertyName = "Name",
                SortCode = 20,
                Row = i + 1,
                Value = list[i].friendlyName,
                Comment = null
            });
        }
    }


    /// <summary>
    /// enumerate all predicates to put data in the existing measurement areas (MA) for the sample
    /// </summary>
    public void GetDatabaseValues_FillPropertiesForCompositions(List<CompositionsForSampleUpdate> compUpdate)
    {
        bool hasFriendlyNameColumnName = dt.Columns.Contains(friendlyNameColumnName);
        for (int rowIdx = 0; rowIdx < dt.Rows.Count; rowIdx++)
        {
            DataRow row = dt.Rows[rowIdx];

            // prepare a predicate for the measurement area
            var pred = new Predicate();
            pred.Properties.Add(new PropertyValue()
            {
                PropertyType = PropertyType.Int,
                PropertyName = "Measurement Area",
                Value = row["Index"]
            });
            var compositionsForSampleUpdate = new CompositionsForSampleUpdate()
            {
                Predicate = pred,
                DeletePreviousProperties = false
            };

            var ids = row[idColumnName].ToString();
            var names = hasFriendlyNameColumnName ? row[friendlyNameColumnName].ToString() : string.Empty;

            // add all phases a a single property "Phases"
            compositionsForSampleUpdate.Properties.Add(new PropertyValue()
            {
                PropertyType = PropertyType.String,
                PropertyName = "Phases",
                SortCode = 5,
                Value = ids
            });

            // iteratively add all phases to the MA as a table PhaseId, Phase Name

            var idsArr = ids.Split(';', StringSplitOptions.RemoveEmptyEntries);
            var namesArr = names.Split(separators, StringSplitOptions.RemoveEmptyEntries);
            if (hasFriendlyNameColumnName)
            {    // idColumnName + friendlyNameColumnName
                var tuples = idsArr.Zip(namesArr, (id, name) => ((id, name))).ToArray();
                for (int i = 0; i < tuples.Length; i++)
                {
                    compositionsForSampleUpdate.Properties.Add(new PropertyValue()
                    {
                        PropertyType = PropertyType.String,
                        PropertyName = "PhaseName",
                        SortCode = 10,
                        Row = i + 1,
                        Value = tuples[i].name
                    });
                    compositionsForSampleUpdate.Properties.Add(new PropertyValue()
                    {
                        PropertyType = PropertyType.String,
                        PropertyName = "Phase",
                        SortCode = 100,
                        Row = i + 1,
                        Value = tuples[i].id
                    });
                }
            }
            else
            {   // idColumnName only
                for (int i = 0; i < idsArr.Length; i++)
                {
                    compositionsForSampleUpdate.Properties.Add(new PropertyValue()
                    {
                        PropertyType = PropertyType.String,
                        PropertyName = "Phase",
                        SortCode = 100,
                        Row = i + 1,
                        Value = idsArr[i]
                    });
                }
            }
            compUpdate.Add(compositionsForSampleUpdate);
        }
    }

}