﻿using Google.Apis.Auth.OAuth2;
using Google.Apis.Services;
using Google.Apis.Sheets.v4;

namespace TypeValidationLibrary;

public class WaferSearchOnSpreadsheet {
    const string range = "B:B"; // range for search (all B column)
    const string lastColumn = "N";    // last column name

    WaferSearchSettings settings;

    /// <summary>
    /// constructor to initialize an object with API credentials and Spreadsheet identifier
    /// </summary>
    /// <param name="credentialsJson">Json with service credentials to Google APIs & Services : Credentials : Service Accounts
    /// C:\RUB\!WORK\!WEB\!Google_Secrets\materialsproject-vic-credentials.json
    /// </param>
    /// <param name="spreadsheetId">Google Spreadsheet identifier</param>
    public WaferSearchOnSpreadsheet(WaferSearchSettings settings)
    {
        this.settings = settings;
    }

    /// <summary>
    /// Search for a specific wafer LaserMark in the Google spreadsheet (as defined in the constructor)
    /// </summary>
    /// <param name="waferLaserMark">wafer LaserMark to search for</param>
    /// <param name="comparator">Func<string, string, bool> comparator. if not set, then default zeroes-triming comparator is used</param>
    /// <returns>Dictionary<string, string> with all key-values pair from the row matched</returns>
    /// <exception cref="Exception">exceptions if no access or wrong document format</exception>
    public async Task<Dictionary<string, string>> Search(string waferLaserMark, Func<string, string, bool>? comparator = null)
    {
        if (comparator == null) {
            comparator = (value, searchValue) => string.CompareOrdinal(value?.TrimStart('0'), searchValue) == 0;
        }
        if (settings.CredentialsJson == null)
        {
            throw new Exception("settings.CredentialsJson == null");
        }
        if (settings.SpreadsheetId == null)
        {
            throw new Exception("settings.SpreadsheetId == null");
        }
        // authorization
        // https://console.cloud.google.com/iam-admin/serviceaccounts/create?inv=1&invt=AbjIzg&project=materialsproject
        var credential = GoogleCredential.FromJson(settings.CredentialsJson)
            //.FromFile(credentialsJsonFile)
            .CreateScoped(SheetsService.Scope.SpreadsheetsReadonly);    // for read-only access

        using (var service = new SheetsService(new BaseClientService.Initializer {
            HttpClientInitializer = credential,
            ApplicationName = "Google Sheets API Wafer Number Reader",
        })) {

            //// optional: check access
            //var spreadsheetRequest = service.Spreadsheets.Get(spreadsheetId);
            //try
            //{
            //    var spreadsheet = await spreadsheetRequest.ExecuteAsync();
            //    Console.WriteLine($"Document opened: {spreadsheet.Properties.Title}");
            //}
            //catch (Google.GoogleApiException ex)
            //{
            //    Console.WriteLine("Error: This is not a Google Table OR access denied.");
            //    Console.WriteLine(ex.Message);
            //    return;
            //}


            // Считываем заголовки столбцов
            var headersRequest = service.Spreadsheets.Values.Get(settings.SpreadsheetId, $"A1:{lastColumn}1");
            var headersResponse = await headersRequest.ExecuteAsync();
            var headers = headersResponse.Values?.FirstOrDefault();

            if (headers == null)
            {
                throw new Exception("No head.");
            }


            // Read data from table
            var request = service.Spreadsheets.Values.Get(settings.SpreadsheetId, range);
            var response = await request.ExecuteAsync();
            var values = response.Values;

            if (values == null || values.Count == 0)
            {
                throw new Exception("Data not found");
            }

            // try to find a row
            for (int row = 1; row <= values.Count; row++)
            {
                // preprocessValue
                if (values[row - 1].Count > 0 && comparator(values[row - 1][0]?.ToString(), waferLaserMark))
                {
                    // if found read the whole row
                    var rowRange = $"A{row}:{lastColumn}{row}"; // suggest that maximum colums up to lastColumn
                    var rowRequest = service.Spreadsheets.Values.Get(settings.SpreadsheetId, rowRange);
                    var rowResponse = await rowRequest.ExecuteAsync();

                    if (rowResponse.Values != null && rowResponse.Values.Count > 0)
                    {
                        var rowData = rowResponse.Values.First();

                        // Формируем словарь "ColumnName - Value"
                        // char colName = 'A'; // Column = (colName++).ToString(), 
                        var result = headers.Zip(rowData, (header, value) => new { Header = header, Value = value })
                                            .ToDictionary(item => item.Header.ToString(), item => item.Value?.ToString());
                        return result;
                    }
                }
            }
            return null;    // Value not found
        }
    }

}
