using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Configuration;
using Microsoft.VisualBasic;
using System.Net.Mime;
using System.Text;
using System.Text.Json.Nodes;
using System.Text.Json;
using TypeCoreLibrary;
using TypeValidationLibrary;
using static System.Runtime.InteropServices.JavaScript.JSType;

namespace WebApiWaferValidator.Controllers;
// https://localhost:7008/WaferValidation/

[ApiController]
[Route("[controller]")]
public class WaferValidationController : ControllerBase
{
    private readonly ILogger<WaferValidationController> logger;
    private readonly IConfiguration config;
    private readonly WaferSearchSettings settings;

    public WaferValidationController(ILogger<WaferValidationController> logger, IConfiguration config)
    {
        this.logger = logger;
        this.config = config;
        settings = WaferSearchOnSpreadsheetWrapper.ReadConfig(config);
    }


    private string ReadInput()
    {
        Request.EnableBuffering();  // important, otherwise can't seek input stream
        Stream req = Request.Body;
        if (req.CanSeek)
        {
            // Reset the position to zero to read from the beginning.
            req.Seek(0, System.IO.SeekOrigin.Begin);
        }
        // string data = Encoding.UTF8.GetString(array, 0, array.Length);
        string data;
        using (var reader = new StreamReader(req, Encoding.UTF8))
        {
            data = reader.ReadToEndAsync().Result;
        }

        return data;
    }

    /// <summary>
    /// Validation of incoming stream from POST body (post call to /EDX/Validation/body)
    /// </summary>
    /// <param name="array">binary request body (POST)</param>
    /// <returns>TypeValidatorResult, specifying validation success</returns>
    [HttpPost("body")]
    [Consumes(MediaTypeNames.Application.Octet)] // application/octet-stream
    [Produces(typeof(TypeValidatorResult))]
    [ProducesResponseType(typeof(TypeValidatorResult), 200)]
    public async Task<TypeValidatorResult> Body([FromBody] byte[] array)
    {
        string data = ReadInput();

        TypeValidatorResult vr;
        int waferId;
        if (!int.TryParse(data, out waferId))
        {
            vr = new TypeValidatorResult(400, "Wafer ID is not an integer number");
            return vr;
        }

        if (waferId < 0)
        {
            vr = new TypeValidatorResult(404, $"Wafer ID is a negative number [{waferId}]");
            return vr;
        }

        if (waferId == 0)
        {
            vr = new TypeValidatorResult() { Warning = "old schema (no lasermark)" };
            return vr;
        }

        // use Google API
        WaferSearchOnSpreadsheet searcher = new WaferSearchOnSpreadsheet(settings);
        string searchValue = waferId.ToString(); // "000015".TrimStart('0');
        var result = await searcher.Search(searchValue, (value, searchValue) =>
            string.CompareOrdinal(value?.TrimStart('0'), searchValue) == 0
        );
        if (result == null)
        {
            Console.WriteLine("Value not found");
            return new TypeValidatorResult(404, $"Wafer ID is not found [{waferId}]");
        }

        var filteredResult = result.Where(x => settings.Keys.Length < 1 || settings.Keys.Contains(x.Key));
        vr = new TypeValidatorResult() { Data = filteredResult.ToList() };
        return vr;
    }
}
